#include <stdio.h>   /* Standard input/output definitions */
#include <stdlib.h>  /* Standard input/output definitions */
#include <string.h>  /* String function definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <time.h>    /* time control definitions */
#include <gtk/gtk.h>    
#include <math.h>

#define ECU_STATUS_NOT_CONNECTED     0
#define ECU_STATUS_CONNECTED         1
#define ECU_STATUS_UNABLE_TO_CONNECT 2


double databuf[140];

int ECU = -1;
int ECUStatus = ECU_STATUS_NOT_CONNECTED;
char ViewRunning = 0;
int timecounter = 0;
unsigned char DID[8] = {0x3, 0x22, 0x1, 0x0, 0x0, 0x0, 0x0, 0x0};
unsigned char RxData[8];
unsigned short Lambda;

GtkWidget *mainwindow;
GtkWidget *mainbox;
GtkWidget *button_connect_controller;
GtkWidget *button_start_view;
GtkWidget *drawing_area_main;
GtkWidget *drawing_area_connect_controller;
guint timeout_id;

gboolean draw_callback_area_main (GtkWidget *widget, cairo_t *cr, gpointer data);
gboolean draw_callback_area_connect_controller (GtkWidget *widget, cairo_t *cr, gpointer data);
gboolean time_handler (gpointer drawing_area);
static void Connect_controller_clicked (GtkWidget *widget, gpointer data);
static void Start_view_clicked (GtkWidget *widget, gpointer data);

void PrintRxData (unsigned char * RxData, unsigned char RxLen);
void PrintTxData (unsigned char * TxData, unsigned char TxLen);
int ReceiveUDSResponse(int ECU, unsigned char * RxData, unsigned char RxLen);
int SendUDSRequest(int ECU, unsigned char * TxData, unsigned char TxLen);
int ConnectECU (char[]);

int main (int argc, char *argv[])
{
    int i;
    for (i = 0; i < 140; i++)
    {
        databuf[i] = 0.0;
    }

    gtk_init(&argc, &argv);

    mainwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size(GTK_WINDOW(mainwindow), 800, 600);
    g_signal_connect (G_OBJECT (mainwindow), "destroy", gtk_main_quit, NULL);

    mainbox = gtk_grid_new();

    button_connect_controller = gtk_button_new_with_label("Connect controller");
    gtk_grid_attach(GTK_GRID(mainbox), button_connect_controller, 0, 0, 1, 1);
    g_signal_connect (button_connect_controller, "clicked", G_CALLBACK (Connect_controller_clicked), NULL);

    drawing_area_connect_controller = gtk_drawing_area_new ();
    gtk_widget_set_size_request (drawing_area_connect_controller, 150, 10);
    gtk_grid_attach(GTK_GRID(mainbox), drawing_area_connect_controller, 1, 0, 1, 1);
    g_signal_connect (G_OBJECT (drawing_area_connect_controller), "draw", G_CALLBACK (draw_callback_area_connect_controller), NULL);


    button_start_view = gtk_button_new_with_label("Start/stop view");
    gtk_grid_attach(GTK_GRID(mainbox), button_start_view, 0, 1, 1, 1);
	g_signal_connect (button_start_view, "clicked", G_CALLBACK (Start_view_clicked), NULL);


    drawing_area_main = gtk_drawing_area_new ();
    gtk_widget_set_size_request (drawing_area_main, 800, 550);
    gtk_grid_attach(GTK_GRID(mainbox), drawing_area_main, 0, 2, 10, 10);
    g_signal_connect (G_OBJECT (drawing_area_main), "draw", G_CALLBACK (draw_callback_area_main), NULL);


    timeout_id = g_timeout_add(50, (GSourceFunc) time_handler, (gpointer) drawing_area_main);

    gtk_container_add(GTK_CONTAINER(mainwindow), mainbox);
    gtk_widget_show_all(mainwindow);

    gtk_main();

    return 0;
}


gboolean draw_callback_area_main (GtkWidget *widget, cairo_t *cr, gpointer data)
{

  cairo_text_extents_t extents;

  double x,y;
  int width, height;
  int i;

  width = gtk_widget_get_allocated_width (widget);
  height = gtk_widget_get_allocated_height (widget);

  cairo_set_source_rgba (cr, 1.0, 1.0, 1.0, 1.0);
  cairo_rectangle(cr, 0, 0, width, height);
  cairo_fill(cr);

  cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
  cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size (cr, 14.0);

  cairo_text_extents (cr, "Lambda", &extents);
  cairo_move_to (cr, 15.0, 250.0);
  cairo_save(cr);
  cairo_rotate(cr, -G_PI/2);
  cairo_show_text (cr, "Lambda");
  cairo_restore(cr);

  char lambda_string[6];
  sprintf(lambda_string, "%d", Lambda);
  cairo_set_source_rgba (cr, 0.5, 0.5, 1.0, 0.5);
  cairo_text_extents (cr, lambda_string, &extents);
  cairo_set_font_size (cr, 150.0);
  cairo_move_to (cr, 250.0, 150.0);
  cairo_show_text (cr, lambda_string);


  cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
  cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size (cr, 14.0);
  cairo_text_extents (cr, "0.6", &extents);
  cairo_move_to (cr, 30.0, 470.0);
  cairo_show_text (cr, "0.6");
  cairo_text_extents (cr, "0.7", &extents);
  cairo_move_to (cr, 30.0, 420.0);
  cairo_show_text (cr, "0.7");
  cairo_text_extents (cr, "0.8", &extents);
  cairo_move_to (cr, 30.0, 370.0);
  cairo_show_text (cr, "0.8");
  cairo_text_extents (cr, "0.9", &extents);
  cairo_move_to (cr, 30.0, 320.0);
  cairo_show_text (cr, "0.9");
  cairo_text_extents (cr, "1.0", &extents);
  cairo_move_to (cr, 30.0, 270.0);
  cairo_show_text (cr, "1.0");
  cairo_text_extents (cr, "1.1", &extents);
  cairo_move_to (cr, 30.0, 220.0);
  cairo_show_text (cr, "1.1");
  cairo_text_extents (cr, "1.2", &extents);
  cairo_move_to (cr, 30.0, 170.0);
  cairo_show_text (cr, "1.2");
  cairo_text_extents (cr, "1.3", &extents);
  cairo_move_to (cr, 30.0, 120.0);
  cairo_show_text (cr, "1.3");
  cairo_text_extents (cr, "1.4", &extents);
  cairo_move_to (cr, 30.0, 70.0);
  cairo_show_text (cr, "1.4");
  cairo_text_extents (cr, "1.5", &extents);
  cairo_move_to (cr, 30.0, 20.0);
  cairo_show_text (cr, "1.5");

  cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
  cairo_move_to (cr, 60, 15);
  cairo_line_to (cr, 60, 465);
  cairo_set_line_width (cr, 1.0);
  cairo_move_to (cr, 761, 15);
  cairo_line_to (cr, 761, 465);
  cairo_stroke (cr);

  cairo_set_source_rgba (cr, 0.8, 1.0, 0.8, 1.0);
  cairo_set_line_width (cr, 30.0);
  cairo_move_to (cr, 61, 265);
  cairo_line_to (cr, 760, 265);
  cairo_stroke (cr);

  cairo_set_source_rgba (cr, 0.9, 0.9, 0.9, 1.0);
  cairo_set_line_width (cr, 1.0);
  for (i = 0; i < 10; i++)
  {
      y = (15 + (i * 50));
      cairo_move_to (cr, 61, y);
      cairo_line_to (cr, 760, y);
  }
  cairo_stroke (cr);

  cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
  cairo_set_line_width (cr, 1.0);
  for (i = 0; i < 139; i++)
  {
      cairo_move_to (cr, (61.0 + (double)(5*i)), databuf[i]);
      cairo_line_to (cr, (66.0 + (double)(5*i)), databuf[i+1]);
  }

  cairo_stroke (cr);
  /*
  char lambda_string[10];
  sprintf(lambda_string, "%i", rxdata);
  cairo_text_extents (cr, lambda_string, &extents);
  cairo_move_to (cr, 300.0, 300.0);
  cairo_show_text (cr, lambda_string);
  */

  return FALSE;
}

gboolean draw_callback_area_connect_controller (GtkWidget *widget, cairo_t *cr, gpointer data)
{
   cairo_text_extents_t extents;
   char message[25];
   double x = 5.0;
   double y = 20.0;

   switch (ECUStatus)
   {
       case ECU_STATUS_CONNECTED:
           strcpy(message, "Connected");
           cairo_set_source_rgba (cr, 0.0, 1.0, 0.0, 1.0);
           break;
       case ECU_STATUS_NOT_CONNECTED:
           strcpy(message, "Not connected");
           cairo_set_source_rgba (cr, 0.0, 0.0, 0.0, 1.0);
           break;
       case ECU_STATUS_UNABLE_TO_CONNECT:
           strcpy(message, "Could not connect");
           cairo_set_source_rgba (cr, 1.0, 0.0, 0.0, 1.0);
           break;
       default:
           break;
   }

   cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
   cairo_set_font_size (cr, 14.0);
   cairo_text_extents (cr, message, &extents);
   cairo_move_to (cr, x,y);
   cairo_show_text (cr, message);

   return FALSE;
}

gboolean time_handler (gpointer drawing_area)
{
    int i;
    if (ViewRunning == 1)
	{
        for (i = 0; i < 139; i++)
        {
            databuf[i] = databuf[i + 1];
        }

        //double Lambda = 0.4 * sin(0.16 * (double)(timecounter));
        //databuf[139] = (((Lambda + 0.4) * 500.0) + 15.0);

        short int rxdata;
        if(ECUStatus == ECU_STATUS_CONNECTED)
        {
            //usleep(50000);
            ReceiveUDSResponse(ECU, RxData, 8);
            rxdata = (RxData[6] | (RxData[7]<<8));
            Lambda = rxdata;
            databuf[139] = (765.0 - ((double)rxdata / 2.0));
            SendUDSRequest(ECU, DID, 8);
        }


	    timecounter++;
	    if(timecounter == 139)
	    {
	        timecounter = 0;
	    }
	    gtk_widget_queue_draw(drawing_area);
	}
	return(TRUE);
}

static void Connect_controller_clicked (GtkWidget *widget, gpointer data)
{
    switch (ECUStatus)
    {
        case ECU_STATUS_NOT_CONNECTED:
        case ECU_STATUS_UNABLE_TO_CONNECT:
            ECU = ConnectECU("WBO2");
            if (ECU > 0)
            {
                ECUStatus = ECU_STATUS_CONNECTED;
                SendUDSRequest(ECU, DID, 8);
            }
            else
            {
                ECUStatus = ECU_STATUS_UNABLE_TO_CONNECT;
            }
            break;
        case ECU_STATUS_CONNECTED:
            close(ECU);
            ECUStatus = ECU_STATUS_NOT_CONNECTED;
            break;
        default:
            break;
    }
	gtk_widget_queue_draw(drawing_area_connect_controller);
}

static void Start_view_clicked (GtkWidget *widget, gpointer data)
{
	if (ViewRunning == 0)
	{
		ViewRunning = 1;
	}
	else
	{
		ViewRunning = 0;
	}
}

int SendUDSRequest(int ECU, unsigned char * TxData, unsigned char TxLen)
{
  int status = 0;
  if (!write(ECU, TxData, TxLen))
  {
    //fprintf(stderr, "Data transmission Error\n");
    close(ECU);
    status = -1;
  }
  return(status);
}

int ReceiveUDSResponse(int ECU, unsigned char * RxData, unsigned char RxLen)
{
  int status = 0;
  if (!read(ECU, RxData, RxLen))
  {
    //fprintf(stderr, "Data reception Error\n");
    close(ECU);
    status = -1;
  }
  return(status);
}

void PrintRxData (unsigned char * RxData, unsigned char RxLen)
{
  unsigned char i;
  for (i = 0; i < RxLen; i++)
  {
	  printf("RxData[%d] = %x\n", i, RxData[i]);
  }
}

void PrintTxData (unsigned char * TxData, unsigned char TxLen)
{
  unsigned char i;
  for (i = 0; i < TxLen; i++)
  {
	  printf("TxData[%d] = %x\n", i, TxData[i]);
  }

}

int ConnectECU (char ECUid[])
{
    int USB = -1;
    char path[] = "/dev/ttyUSB0";
    unsigned char i;
    unsigned char loc_ECUConnected = 0;

    //printf("Searching for %s...\n", ECUid);
    for (i = 0; ((i < 10) && (loc_ECUConnected == 0)); i++)
    {
        path[11] = i + 48;
        //printf("Trying to connect at %s... ", path);
        USB = open(path, O_RDWR| O_NOCTTY);
        if (USB >= 0)
        {
	        struct termios tty;
	        memset (&tty, 0, sizeof tty);
	        tcgetattr(USB, &tty);
	        cfsetospeed(&tty,(speed_t)B9600);
	        cfsetispeed(&tty,(speed_t)B9600);
	        tty.c_iflag &= ~IGNBRK;
	        tty.c_lflag = 0;
	        tty.c_oflag = 0;
	        tty.c_iflag &= ~(IXON | IXOFF | IXANY);
	        tty.c_cflag &= ~(PARENB | PARODD);
	        tty.c_cflag     &=  ~CSTOPB;
	        tty.c_cflag     &=  ~CSIZE;
	        tty.c_cflag     |=  CS8;
	        tty.c_cflag     &=  ~CRTSCTS;       // no flow control
	        tty.c_cc[VMIN]      =   8;          // read doesn't block
	        tty.c_cc[VTIME]     =   5;          // 0.5 seconds read timeout
	        tty.c_cflag     |=  (CREAD | CLOCAL); // turn on READ & ignore ctrl lines
            cfmakeraw(&tty);
	        tcflush(USB, TCIFLUSH);
            tcsetattr(USB, TCSANOW, &tty);

            unsigned char TxData[8] = {0x03, 0x22, 0x0, 0x1, 0x0, 0x0, 0x0, 0x0};
            unsigned char RxData[8];
            SendUDSRequest(USB, TxData, 8);
            usleep(30000);
            ReceiveUDSResponse(USB, RxData, 8);
            //PrintRxData(RxData, 8);
            if ((RxData[0] == 0x7) && (RxData[1] == 0x62) && (RxData[2] == 0x0) && (RxData[3] == 0x1)
                && (RxData[4] == ECUid[0]) && (RxData[5] == ECUid[1]) && (RxData[6] == ECUid[2]) && (RxData[7] == ECUid[3]))
            {
                loc_ECUConnected = 1;
                //printf("%s connected at %s\n", ECUid, path);
            }
            else
            {
                //printf("Device connected at %s is not %s\n", path, ECUid);
                //PrintTxRxData(TxData, RxData);
                close(USB);
                USB = (-1);
            }
        }
		else
		{
			//printf("No RS-232 interface fount at %s\n", path);
			close(USB);
			USB = (-1);
		}
    }
    return(USB);
}
